#include <string.h>

#include "http_loop.h"
#include "log.h"

http_parse_error http_loop_parse_request(event_loop_conn *conn) {
  http_loop_ctx *ctx = conn->ctx;

  // First we try to parse the incoming HTTP request
  size_t num_headers = HTTP_MAX_ALLOWED_HEADERS;
  http_request *req = &ctx->req;

  int res =
      phr_parse_request((const char *)&conn->rbuf[conn->rbuf_read],
                        conn->rbuf_size - conn->rbuf_read, &req->method,
                        &req->method_len, &req->path, &req->path_len,
                        &req->minor_version, req->headers, &num_headers, 0);

  if (res == -1) {
    return http_parse_error_invalid;
  } else if (res == -2) {
    return http_parse_error_incomplete;
  }

  req->len = res;

  // Split path into path & query
  size_t i = 0;
  bool no_query = true;

  while (no_query && i < req->path_len) {
    if (req->path[i] == '?') {
      // Ensure we don't store an invalid pointer if the request simply ends
      // with '?'
      if (i + 1 < req->path_len) {
        req->query = &req->path[i + 1];
        req->query_len = req->path_len - (i + 1);
      }

      req->path_len = i;

      no_query = false;
    }

    i++;
  }

  // Ensure we clear the old request's query
  if (no_query) {
    req->query = NULL;
    req->query_len = 0;
  }

  return http_parse_error_ok;
}

void http_loop_process_request(event_loop_conn *conn) {
  http_loop_ctx *ctx = conn->ctx;

  // The end of the list of steps is marked with a NULL. When this is reached,
  // we simply set the route to NULL, letting the http loop know a new request
  // can be handled.
  if (ctx->route->steps[ctx->current_step] == NULL) {
    ctx->route = NULL;
    ctx->current_step = 0;
    return;
  }

  // We execute the next step
  // NOTE can/should we execute more than one step at once here?
  if (ctx->route->steps[ctx->current_step](conn)) {
    ctx->current_step++;
  }
}

bool http_loop_route_request(event_loop_conn *conn) {
  http_loop_ctx *ctx = conn->ctx;
  http_loop_gctx *gctx = ctx->g;

  info("%.*s %.*s", ctx->req.method_len, ctx->req.method, ctx->req.path_len,
       ctx->req.path);

  http_route *route;

  for (size_t i = 0; i < gctx->route_count; i++) {
    route = &gctx->routes[i];

    switch (route->type) {
    case http_route_literal:
      if (strncmp(route->path, ctx->req.path, ctx->req.path_len) == 0) {
        ctx->route = route;
        return true;
      }
    // TODO
    case http_route_regex:;
    }
  }

  // Fallthrough is to write a 404
  http_write_standard_response(conn, http_not_found);

  return false;
}
